from Data import Data

data = Data(selection=["australian", "spambase"])
datasets = data.load()
print(len(datasets))

data = Data(selection=("all", ["balanced", "binary"]))
datasets = data.load()
print(len(datasets))

data = Data(selection=("any", ["balanced", "binary"]))
datasets = data.load()
print(len(datasets))



from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import DistanceMetric
import numpy as np
import math


class ExposerClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, radius=None, p=2, memory=250):
        self.radius = radius
        self.p = p
        self.memory = memory

    def fit(self, X, y):
        self.scaler = StandardScaler().fit(X)
        self.X_ = self.scaler.transform(X)
        self.y_ = np.copy(y)
        self.classes_, self.prior = np.unique(y, return_counts=True)
        self.metric = DistanceMetric.get_metric(metric="euclidean")
        return self

    def partial_fit(self, X, y, classes=None):
        self.scaler = StandardScaler().fit(X)
        self.X_ = (
            np.concatenate((self.X_, self.scaler.transform(X)), axis=0)
            if hasattr(self, "X_")
            else self.scaler.transform(X)
        )
        self.y_ = (
            np.concatenate((self.y_, y), axis=0) if hasattr(self, "y_") else np.copy(y)
        )

        self.classes_ = classes
        _, self.prior = np.unique(y, return_counts=True)
        if self.classes_ is None:
            self.classes_, self.prior = np.unique(y, return_counts=True)

        self.metric = DistanceMetric.get_metric(metric="euclidean")

        if self.X_.shape[0] > self.memory:
            self.X_, self.y_ = self.X_[-self.memory :, :], self.y_[-self.memory :]
        return self

    def predict_proba(self, X):
        X_ = self.scaler.transform(X)
        if self.radius is None:
            self.radius = np.sum(np.std(X_, axis=0))

        distances = self.metric.pairwise(X_, self.X_)
        return np.array(
            [
                np.sum((self.radius - distances[:, self.y_ == label]).clip(0), axis=1)
                / self.prior[label]
                for label in self.classes_
            ]
        )

    def predict(self, X):
        return np.argmax(self.predict_proba(X), axis=0)

    def minkowski_distance(self, x, y):
        diff = np.fabs(x - y)
        diff_p = diff ** self.p
        return max(math.pow(np.sum(diff_p), 1 / self.p), np.nextafter(0, 1))
        # return 1.0



from sklearn.datasets import make_classification
import sklearn as skl
import weles as ws


def multiclass_dataset():
    return make_classification(random_state=1410, n_classes=5,
                               n_informative=11)


def test_balanced_accuracy_score():
    clf = ws.classifiers.LinearClassifier()
    X, y = multiclass_dataset()
    clf.fit(X, y)
    y_pred = clf.predict(X)
    score_1 = ws.metrics.balanced_accuracy_score(y, y_pred)
    score_2 = skl.metrics.balanced_accuracy_score(y, y_pred)
    assert score_1 == score_2



from .Cluens import Cluens
from .Geometron import Geometron
from .GES import GES
from .NURS import NURS
from .RandomSubspace import RandomSubspace
from .UMCE import UMCE

__all__ = ["Cluens", "Geometron", "GES", "NURS", "RandomSubspace", "UMCE"]



from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import DistanceMetric
import numpy as np
import math


class RandomSubspace(BaseEstimator):
    def __init__(self, base_estimator, n_members=100, subspace_size=3):
        self.base_estimator = base_estimator
        self.n_members = n_members
        self.subspace_size = subspace_size

    def fit(self, X, y):
        self.n_features = X.shape[1]
        self.subspaces = np.random.randint(
            self.n_features, size=(self.n_members, self.subspace_size)
        )

        self.ensemble = []
        for subspace in self.subspaces:
            clf = clone(self.base_estimator).fit(X[:, subspace], y)
            self.ensemble.append(clf)

        return self

    def predict_proba(self, X):
        esm = np.mean(
            np.array(
                [
                    clf.predict_proba(X[:, self.subspaces[i]])
                    for i, clf in enumerate(self.ensemble)
                ]
            ),
            axis=0,
        )

        return esm.T

    def predict(self, X):
        pp = self.predict_proba(X)

        y_pred = np.argmax(pp, axis=1)

        return y_pred



import numpy as np

def entropy_measure_e(ensemble, X, y):
    L = len(ensemble)
    return np.mean(
        (
            L // 2
            - np.abs(
                np.sum(
                    y[np.newaxis, :] == np.array([clf.predict(X) for clf in ensemble]),
                    axis=0,
                )
                - L // 2
            )
        )
        / (L / 2)
    )



"""
Class description

Date:
Authors:
"""

# imports
import numpy as np
from sklearn.model_selection import RepeatedStratifiedKFold
from sklearn.base import clone
from tabulate import tabulate
from tqdm import tqdm
from scipy.stats import rankdata

VERBOSE_COLUMNS = 80


class Evaluator():
    def __init__(self, datasets, protocol=(1, 5, None)):
        self.datasets = datasets
        self.protocol = protocol

    def process(self, clfs, verbose=False):
        """
        This function is used to process declared evaluation protocol
        through all given datasets and classifiers.
        It results with stored predictions and corresponding labels.

        Input atguments description:
        clfs: dictonary that contains estimators names and objects
              ["name"] : obj
        """
        self.clfs = clfs

        # Establish protocol
        self.m, self.k, self.random_state = self.protocol

        skf = RepeatedStratifiedKFold(n_splits=self.k, n_repeats=self.m,
                                      random_state=self.random_state)
        self.predictions = np.zeros([len(self.datasets), len(self.clfs),
                                     self.m * self.k], dtype=object)
        self.true_values = np.zeros([len(self.datasets), self.m * self.k],
                                    dtype=object)

        # Iterate over datasets
        for dataset_id, dataset_name in enumerate(tqdm(self.datasets,
                                                       desc="DTS",
                                                       ascii=True,
                                                       disable=not verbose)):
            X, y = self.datasets[dataset_name]
            for fold_id, (train, test) in enumerate(skf.split(X, y)):
                self.true_values[dataset_id, fold_id] = y[test]
                for clf_id, clf_name in enumerate(self.clfs):
                    clf = clone(self.clfs[clf_name])
                    clf.fit(X[train], y[train])
                    y_pred = clf.predict(X[test])
                    self.predictions[dataset_id, clf_id, fold_id] = y_pred

        return self

    def score(self, metrics, verbose=False, return_flatten=True):
        """
        description

        Input arguments description:
        metrics: dictonary that contains metrics names and functions
                 ["name"] : function
        """
        self.metrics = metrics

        # Prepare storage for scores
        # DB x CLF x FOLD x METRIC
        self.scores = np.array([[[[
            metrics[m_name](
                self.predictions[db_idx, clf_idx, f_idx],
                self.true_values[db_idx, f_idx])
            for m_name in self.metrics]
            for f_idx in range(self.m * self.k)]
            for clf_idx, clf in enumerate(self.clfs)]
            for db_idx, db_name in enumerate(self.datasets)])

        # Store mean scores and stds
        # DB x CLF x METRIC
        self.mean_scores = np.mean(self.scores, axis=2)
        self.stds = np.std(self.scores, axis=2)

        # Verbose mode
        if verbose:
            lmn = len(max(list(self.metrics.keys()), key=len))
            lmc = (VERBOSE_COLUMNS-lmn)//2
            for m, metric in enumerate(self.metrics):
                print(lmc*"#", metric.center(lmn), lmc*"#")

                scores_ = self.mean_scores[:, :, m]

                # ranks
                ranks = []
                for row in scores_:
                    ranks.append(rankdata(row).tolist())
                ranks = np.array(ranks)
                mean_ranks = np.mean(ranks, axis=0)

                names_column = np.array(list(self.datasets.keys())).reshape(
                    len(self.datasets), -1)
                scores_table = np.concatenate((names_column, scores_), axis=1)
                print(tabulate(scores_table, headers=self.clfs.keys(),
                               floatfmt=".3f"))

                print(lmc*"-", "Mean ranks".center(lmn), lmc*"-")
                print(tabulate(mean_ranks[np.newaxis, :],
                               headers=self.clfs.keys(), floatfmt=".3f"))

        # Give output
        return {
            True: (self.mean_scores, self.stds),
            False: self.scores,
        }[return_flatten]



from sklearn.base import ClassifierMixin, BaseEstimator
import numpy as np
from sklearn.decomposition import PCA
from sklearn.svm import SVC


class PCASSE(ClassifierMixin, BaseEstimator):
    def __init__(self, n_components=4, subspace_size=4):
        self.n_components = n_components
        self.subspace_size = subspace_size

    def fit(self, X, y, classes=None):
        # Calculate PCA components
        components = np.abs(
            PCA(n_components=self.n_components, svd_solver="full").fit(X).components_
        )

        # Gather ensemble
        self.subspaces = np.array(
            [np.argsort(-row)[: self.subspace_size] for row in components]
        )

        # Build ensemble
        self.ensemble = [SVC().fit(X[:, subspace], y) for subspace in self.subspaces]

        return self

    def predict(self, X):
        return (
            np.mean(
                np.array(
                    [
                        self.ensemble[i].decision_function(X[:, subspace])
                        for i, subspace in enumerate(self.subspaces)
                    ]
                ),
                axis=0,
            )
            > 0
        ).astype(int)


class PCASSEE(ClassifierMixin, BaseEstimator):
    def __init__(self, distribuant_treshold=0.1, subspace_size=4):
        self.distribuant_treshold = distribuant_treshold
        self.subspace_size = subspace_size

    def fit(self, X, y, classes=None):
        # Calculate PCA components

        pca = PCA(svd_solver="full").fit(X)
        components = np.abs(pca.components_)

        # Z EVR
        evrd = np.add.accumulate(pca.explained_variance_ratio_)
        self.n_components = np.where(evrd > self.distribuant_treshold)[0][0]

        if self.n_components == 0:
            self.n_components = 1

        # print(evrd)
        # print("%i COMPONENTS" % self.n_components)
        components = components[: self.n_components, :]

        # Calculate subspace size
        self.subspace_size = 4

        # Gather ensemble
        self.subspaces = np.array(
            [np.argsort(-row)[: self.subspace_size] for row in components]
        )

        # Build ensemble
        self.ensemble = [SVC().fit(X[:, subspace], y) for subspace in self.subspaces]

        return self

    def predict(self, X):
        return (
            np.mean(
                np.array(
                    [
                        self.ensemble[i].decision_function(X[:, subspace])
                        for i, subspace in enumerate(self.subspaces)
                    ]
                ),
                axis=0,
            )
            > 0
        ).astype(int)


class RS(ClassifierMixin, BaseEstimator):
    def __init__(self, n_estimators=20, subspace_size=4):
        self.n_estimators = n_estimators
        self.subspace_size = subspace_size

    def fit(self, X, y, classes=None):
        # Calculate PCA components

        pca = PCA(svd_solver="full").fit(X)
        components = np.abs(pca.components_)

        # Gather ensemble
        self.subspaces = np.random.randint(
            X.shape[1], size=(self.n_estimators, self.subspace_size)
        )

        # Build ensemble
        self.ensemble = [SVC().fit(X[:, subspace], y) for subspace in self.subspaces]

        return self

    def predict(self, X):
        return (
            np.mean(
                np.array(
                    [
                        self.ensemble[i].decision_function(X[:, subspace])
                        for i, subspace in enumerate(self.subspaces)
                    ]
                ),
                axis=0,
            )
            > 0
        ).astype(int)



"""
Subspaced Gaussian Naive Bayes
"""
from sklearn.base import ClassifierMixin, BaseEstimator
import numpy as np


class SSGNB(ClassifierMixin, BaseEstimator):
    def __init__(self):
        pass

    def fit(self, X, y, classes=None):
        # Gather prior info
        self.classes, self.prior = np.unique(y, return_counts=True)
        self.prior = self.prior / X.shape[1]
        self.n_features = X.shape[1]
        self.n_classes = len(self.classes)

        # Calculate current norm
        self.means = np.array(
            [np.mean(X[y == label], axis=0) for label in self.classes]
        )
        self.stds = np.array([np.std(X[y == label], axis=0) for label in self.classes])

        return self

    def predict(self, X, subspaces=None):
        if subspaces is None:
            # Calculate distribution density
            ps = np.product(
                [self._pdf(X, self.stds[c], self.means[c]).T for c in self.classes],
                axis=1,
            )

            # Establish and return prediction
            y_pred = np.argmax(self.prior[:, np.newaxis] * ps, axis=0)
            return y_pred
        else:
            # One estimation to meet all requirements
            # To jest fantastyczny tensor z niezależnymi wartościami
            # funkcji gęstości rozkładu dla predykowanego zbioru.
            # Ma wymiary:
            # - klasa
            # - cecha
            # - testowany wzorzec
            # [gęstości rozkładu]
            psf = np.array(
                [self._pdf(X, self.stds[c], self.means[c]).T for c in self.classes]
            )

            # Możemy sobie teraz przeiterować podprzestrzenie i dla każdej z
            # nich wyznaczyć podprzestrzenny produkt. Otrzymamy tensor pse o
            # wymiarach:
            # - identyfikator podprzestrzeni
            # - klasa
            # - testowany wzorzec
            # [gęstości rozkładu]
            pse = np.array([np.product(psf[:, ss, :], axis=1) for ss in subspaces])

            # Z takiego cudu możliwa jest to prostego wyliczenia macierz
            # predykcji komitetu o wymiarach:
            # - identyfikator podprzestrzeni
            # - predykcja
            # [predykcje]
            y_preds = np.argmax(self.prior[np.newaxis, :, np.newaxis] * pse, axis=1)

            return y_preds

    def _pdf(self, x, std=1, mean=0):
        """
        Probability density function.
        """
        return np.exp(-(0.5) * np.power((x - mean) / std, 2)) / (
            std * np.sqrt(2 * np.pi)
        )



from sklearn.base import ClassifierMixin, BaseEstimator, clone
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier
import numpy as np



class NURS(BaseEstimator, ClassifierMixin):
    """
    Non-uniform random subspace
    For drifted data streams
    """
    def __init__(self, best=False, rs=False, n_f=2, n=20, ensemble=False):
        self.coefs = []
        self.best = best
        self.rs = rs
        self.n_f = n_f
        self.n = n  # number of subspaces
        self.ensemble = ensemble

        self.goods = []
        self.bads = []
        np.random.seed(123)

    def partial_fit(self, X, y, classes=None):
        self._X = np.copy(X)
        self._y = np.copy(y)
        n_features = X.shape[1]

        self.scaler = StandardScaler()
        rescaled_X = self.scaler.fit_transform(self._X)
        self.clf = LogisticRegression(solver="lbfgs", multi_class="auto")
        self.clf.fit(rescaled_X, y)

        coef = np.copy(np.abs(self.clf.coef_[0]))
        coef -= np.min(coef)
        coef /= np.sum(coef)
        self.coefs.append(coef)

        self.good = np.argsort(-coef)[: self.n_f]
        self.bad = np.argsort(coef)[: self.n_f]

        self.goods.append(self.good)
        self.bads.append(self.bad)

        # Train
        # Ensemble
        if self.ensemble:
            if self.best:
                self.good_subspaces = np.random.choice(
                    list(range(n_features)), size=(self.n, self.n_f), p=coef
                )
                self.good_ensemble = [
                    DecisionTreeClassifier().fit(X[:, subspace], y)
                    for subspace in self.good_subspaces
                ]
            if not self.best:
                bcoef = np.copy(coef)
                bcoef = 1 - bcoef
                bcoef -= np.min(bcoef)
                bcoef /= np.sum(bcoef)

                self.bad_subspaces = np.random.choice(
                    list(range(n_features)), size=(self.n, self.n_f), p=bcoef
                )

                self.bad_ensemble = [
                    DecisionTreeClassifier().fit(X[:, subspace], y)
                    for subspace in self.bad_subspaces
                ]
            if self.rs:
                self.random_subspaces = np.random.choice(
                    list(range(n_features)), size=(self.n, self.n_f)
                )

                self.random_ensemble = [
                    DecisionTreeClassifier().fit(X[:, subspace], y)
                    for subspace in self.random_subspaces
                ]
        else:
            if self.rs:
                self.rand = np.random.randint(X.shape[1], size=self.n_f)
                self.clf_rand = DecisionTreeClassifier()
                self.clf_rand.fit(X[:, self.rand], y)
            if self.best:
                self.clf_good = DecisionTreeClassifier()
                self.clf_good.fit(X[:, self.good], y)
            if not self.best:
                self.clf_bad = DecisionTreeClassifier()
                self.clf_bad.fit(X[:, self.bad], y)

    def predict(self, X):
        if self.ensemble:
            if self.rs:
                esm = np.array(
                    [
                        clf.predict_proba(X[:, self.random_subspaces[i]])
                        for i, clf in enumerate(self.random_ensemble)
                    ]
                )
                fesm = np.sum(esm, axis=0)
                y_pred = np.argmax(fesm, axis=1)

                return y_pred
            if self.best:
                esm = np.array(
                    [
                        clf.predict_proba(X[:, self.good_subspaces[i]])
                        for i, clf in enumerate(self.good_ensemble)
                    ]
                )
                fesm = np.sum(esm, axis=0)
                y_pred = np.argmax(fesm, axis=1)
                return y_pred
            if not self.best:
                esm = np.array(
                    [
                        clf.predict_proba(X[:, self.bad_subspaces[i]])
                        for i, clf in enumerate(self.bad_ensemble)
                    ]
                )
                fesm = np.sum(esm, axis=0)
                y_pred = np.argmax(fesm, axis=1)

                return y_pred
        else:
            if self.rs:
                set = X[:, self.rand]
                return self.clf_rand.predict(set)

            if self.best:
                set = X[:, self.good]
                return self.clf_good.predict(set)
            if not self.best:
                set = X[:, self.bad]
                return self.clf_bad.predict(set)


class SDTC(BaseEstimator, ClassifierMixin):
    def __init(self):
        pass

    def partial_fit(self, X, y, classes=None):
        self.clf = DecisionTreeClassifier()
        self.clf.fit(X, y)

    def predict(self, X):
        return self.clf.predict(X)



import numpy as np
from scipy import stats
from sklearn.base import clone, ClassifierMixin, BaseEstimator
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler

RULES = ["mean", "gmean", "hmean", "rbh"]
KUNCHEVA = 0.000001


class Geometron(ClassifierMixin, BaseEstimator):
    def __init__(
        self, base_estimator, n_estimators=3, random_state=None, rule="gmean", sigma=3
    ):
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.random_state = random_state
        self.rule = rule
        self.sigma = sigma

    def fit(self, X, y):
        np.random.seed(self.random_state)
        self.classes_ = np.unique(y)
        self.n_samples, self.n_features = X.shape
        self.means = []
        self.ensemble = []
        for i in range(self.n_estimators):
            sel = np.random.randint(self.n_samples, size=int(self.n_samples))
            self.ensemble.append(clone(self.base_estimator).fit(X[sel], y[sel]))

        # Test scale for the needs of geometric means
        decfuncs = np.array(
            [clf.decision_function(X) for clf in self.ensemble]
        ).T.reshape(-1)
        self.std = np.sqrt(np.sum(np.power(decfuncs, 2), axis=0) / len(decfuncs))

        return self

    def decfunc(self, X):
        decfuncs = np.array([clf.decision_function(X) for clf in self.ensemble])
        s_decfuncs = decfuncs / self.std
        s_decfuncs += self.sigma
        s_decfuncs = s_decfuncs / (self.sigma * 2)
        s_decfuncs = np.clip(s_decfuncs, KUNCHEVA, 1 - KUNCHEVA)

        if self.rule == "gmean":
            decfunc = (stats.gmean(s_decfuncs, axis=0) - 0.5) * self.sigma * 2
        elif self.rule == "hmean":
            decfunc = (stats.hmean(s_decfuncs, axis=0) - 0.5) * self.sigma * 2
        elif self.rule == "mean":
            decfunc = np.mean(decfuncs, axis=0)
        elif self.rule == "rbh":
            # decfunc = np.mean(decfuncs, axis=0)
            gmin = np.min(decfuncs, axis=0)
            gmax = np.max(decfuncs, axis=0)
            gmed = np.median(decfuncs, axis=0)

            h1 = gmax - gmed
            h2 = gmed - gmin

            hm = stats.hmean([h1, h2], axis=0)
            ghm = np.zeros((gmin.shape))
            mask = gmed > (gmax + gmin)
            mask2 = gmed <= (gmax + gmin)
            ghm[mask] = gmax[mask] - hm[mask]
            ghm[mask2] = gmin[mask2] + hm[mask2]

            decfunc = ghm

        return decfunc

    def predict(self, X):
        decfunc = self.decfunc(X)

        y_pred = decfunc > 0

        return y_pred



from sklearn.datasets import make_classification
import weles as ws


def dataset():
    return make_classification(random_state=1410)


def test_geometron():
    base_estimator = ws.classifiers.LinearClassifier()
    clf = ws.ensembles.Geometron(base_estimator)
    X, y = dataset()
    clf.fit(X, y)
    y_pred = clf.predict(X)



import numpy as np


def balanced_accuracy_score(y_true, y_pred):
    classes = np.unique(y_true)
    n_classes = classes.shape[0]
    _ = np.arange(n_classes*n_classes).reshape(n_classes, n_classes)
    _ = (_ % n_classes, _ // n_classes)
    cm = np.sum((y_true[:, np.newaxis] == classes[np.newaxis, :]).T[_[0]] *
                (y_pred[:, np.newaxis] == classes[np.newaxis, :]).T[_[1]],
                axis=2)
    return np.mean(cm.diagonal() / np.sum(cm, axis=0))



"""
Regression-aided optimizer.
"""

from sklearn.base import BaseEstimator, clone
from sklearn.model_selection import KFold
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import explained_variance_score
import numpy as np
from scipy.optimize import minimize, rosen, rosen_der


class RAO(BaseEstimator):
    def __init__(
        self,
        base_estimator,
        n_splits=5,
        n_guesses=100000,
        metric=explained_variance_score,
        random_state=None,
        bounds=[(1201, 3500), (601, 1200), (0, 600)],
    ):
        self.base_estimator = base_estimator
        self.n_splits = n_splits
        self.metric = metric
        self.random_state = random_state
        self.n_guesses = n_guesses
        self.bounds = bounds

    def fit(self, X, y):
        # Store features and labels
        self.X = np.copy(X)
        self.y = np.copy(y)

        # Prepare storage for ensemble parameters
        self.ensemble = []
        self.feature_scalers = []
        self.label_scalers = []
        self.scores = np.zeros(self.n_splits)

        # Build regression model
        kfold = KFold(
            n_splits=self.n_splits, shuffle=True, random_state=self.random_state
        )
        for fold, (train, test) in enumerate(kfold.split(X)):
            # Train scalers
            feature_scaler = StandardScaler().fit(X[train])
            label_scaler = StandardScaler().fit(y[train].reshape(-1, 1))

            # Build member model
            reg = clone(self.base_estimator)
            reg.fit(
                feature_scaler.transform(X[train]),
                label_scaler.transform(y[train].reshape(-1, 1)),
            )

            # Test member model
            y_pred = reg.predict(feature_scaler.transform(X[test]))
            scaled_y_test = label_scaler.transform(y[test].reshape(-1, 1))
            score = self.metric(scaled_y_test, y_pred)

            # Store member model
            self.ensemble.append(reg)
            self.feature_scalers.append(feature_scaler)
            self.label_scalers.append(label_scaler)
            self.scores[fold] = score

        # Normalize scores to weights
        self.weights = self.scores / np.sum(self.scores)

        # Estimate optimization bounds
        # min_bounds = np.min(X, axis=0)
        # max_bounds = np.max(X, axis=0)
        # self.bounds = [(min_bounds[i], max_bounds[i]) for i in range(X.shape[1])]

        return self

    def predict(self, X):
        # Acquire all predictions
        y_pred_ensemble = np.array(
            [
                reg.predict(self.feature_scalers[i].transform(X))
                for i, reg in enumerate(self.ensemble)
            ]
        )

        if len(y_pred_ensemble.shape) == 3:
            y_pred_ensemble = y_pred_ensemble[:, :, 0]
        print("YPREDENS", y_pred_ensemble.shape)

        # print(y_pred_ensemble.shape)

        # Rescale them
        rescaled_y_pred_ensemble = np.array(
            [
                self.label_scalers[i].inverse_transform(y_pred)
                for i, y_pred in enumerate(y_pred_ensemble)
            ]
        )
        """ FORMER
        rescaled_y_pred_ensemble = np.mean(
            np.array(
                [
                    self.label_scalers[i].inverse_transform(y_pred)
                    for i, y_pred in enumerate(y_pred_ensemble)
                ]
            ),
            axis=2,
        )
        """
        # print(rescaled_y_pred_ensemble.shape)

        # print("REAL", self.y[:3], self.y[-3:])
        # print("RESC", rescaled_y_pred_ensemble, rescaled_y_pred_ensemble.shape)

        # Weight and flattern them
        y_pred = np.sum(rescaled_y_pred_ensemble * self.weights[:, np.newaxis], axis=0)

        return y_pred

    def _optfun(self, x):
        # y =
        # print("OPTFUN", x, y)
        return self.predict([x])

    def optimize(self):
        """
        Monte Carlo optimizer.
        """
        np.random.seed(self.random_state)
        # Generate guesses
        X = np.array(
            [
                np.random.randint(bound[0], bound[1], self.n_guesses)
                for bound in self.bounds
            ]
        ).T

        # Calculate predictions
        y_pred = self.predict(X)

        # Gather minimized prediction
        min_idx = np.argmin(y_pred)

        return (X[min_idx], y_pred[min_idx])

        print(X, X.shape)
        print(y_pred, y_pred.shape)

        exit()
        # res = minimize(self._optfun, [self.stored_minimum()[0]], bounds=self.bounds)
        # return (res.x.astype(int), res.fun[0])

    def inner_confidence(self):
        return np.mean(self.scores)

    def stored_minimum(self):
        """
        Returns x and y for stored minimum, supplemented with predicted y.
        """
        min_idx = np.argmin(self.y)
        y_pred = self.predict([self.X[min_idx]])[0]
        return (self.X[min_idx].astype(int), y_pred, self.y[min_idx])



import numpy as np
from scipy import stats


def t_test_corrected(a, b, J=5, k=5):
    """
    Corrected t-test for repeated cross-validation.
    input, two 2d arrays. Repetitions x folds
    As default for 5x5CV
    """
    if J*k != a.shape[0]:
        raise Exception('%i scores received, but J=%i, k=%i (J*k=%i)' % (
            a.shape[0], J, k, J*k
        ))

    d = a - b
    bar_d = np.mean(d)
    bar_sigma_2 = np.var(d.reshape(-1), ddof=1)
    bar_sigma_2_mod = (1 / (J * k) + 1 / (k - 1)) * bar_sigma_2
    t_stat = bar_d / np.sqrt(bar_sigma_2_mod)
    pval = stats.t.sf(np.abs(t_stat), (k * J) - 1) * 2
    return t_stat, pval


def t_test_13(a, b, corr=0.6):
    """
    Corrected t-test for repeated cross-validation.
    input, two 2d arrays. Repetitions x folds
    """
    k = len(a)  # J - repetitions, k - folds
    d = a - b
    bar_d = np.mean(d)
    bar_sigma_2 = np.var(d.reshape(-1), ddof=1)
    bar_sigma_2_mod = (1 / (k * (1 - corr))) * bar_sigma_2
    t_stat = bar_d / np.sqrt(bar_sigma_2_mod)
    pval = stats.t.sf(np.abs(t_stat), k - 1) * 2
    return t_stat, pval


def t_test_rel(a, b):
    """
    Paired, relative t-test.
    """
    J = len(a)
    d = a - b
    bar_d = np.mean(d)
    bar_sigma_2 = np.var(d.reshape(-1), ddof=1)
    bar_sigma_2_mod = (1 / J) * bar_sigma_2
    t_stat = bar_d / np.sqrt(bar_sigma_2_mod)
    pval = stats.t.sf(np.abs(t_stat), J - 1) * 2
    return t_stat, pval


IMPLEMENTED_TESTS = {
    "t_test_corrected": (t_test_corrected, {'J': 5, 'k': 5}),
    "t_test_13": (t_test_13, {'corr': 0.6}),
    "t_test_rel": (t_test_rel, {})
}



"""
Genetic Ensemble Selection
"""

from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import f1_score
import numpy as np
from scipy.special import binom

VAR = 10000000
np.set_printoptions(suppress=True, precision=3)


class GES(BaseEstimator, ClassifierMixin):
    """
    Genetic Ensemble Selection
    """

    def __init__(
        self,
        pool_size=20,
        ensemble_size=5,
        num_iter=50,
        base_clf=LogisticRegression(solver="lbfgs"),
        elite_limit=1,
        p_crossing=0.025,
        p_mutation=0.01,
        alpha=0,
        beta=0,
        metric=f1_score,
        random_state=13,
    ):
        # Ensemble parameters
        self.pool_size = pool_size  # Number of ensembles in pool
        self.ensemble_size = ensemble_size  # Number of classifiers in ensemble
        self.base_clf = base_clf

        # Genetic parameters
        self.num_iter = num_iter
        self.elite_limit = elite_limit
        self.p_crossing = p_crossing
        self.p_mutation = p_mutation
        self.random_state = random_state

        # Regularization parameters
        self.alpha = alpha  # strength of parameters usage
        self.beta = beta  # strength of hamming distance
        self.metric = metric

    # @profile
    def fit(self, X, y):
        """
        Process training set.
        """
        # Setting random state
        np.random.seed(self.random_state)

        # Store training set properties
        self.X_, self.y_ = X, y
        self.n_features = X.shape[1]
        self.n_objects = X.shape[0]
        self.classes = np.unique(y)
        self.n_classes = len(self.classes)

        # Build base model on all features and store default coef
        self.base_clf.fit(self.X_, self.y_)
        if hasattr(self.base_clf, "coef_"):
            self.coef_ = np.copy(self.base_clf.coef_[0])
        if hasattr(self.base_clf, "sigma_"):
            self.sigma_ = np.copy(self.base_clf.sigma_)

        # Prepare storage for processing
        self.all_scores = []
        self.all_qualities = []

        # Prepare random feature selection model
        self.model_ = (
            np.random.randint(
                2, size=(self.pool_size, self.ensemble_size, self.n_features)
            )
            == 1
        )

        # Prepare probas storage
        self.probas_storage = {}

        # Iterate epochs
        for i in range(self.num_iter):
            # Perform mutation
            self.mutate()

            # Perform crossing
            self.cross()

            # Get model probas, calculate supports and return scores
            probas = self.get_model_probas()
            supports = np.mean(probas, axis=1)
            y_preds = np.argmax(supports, axis=2)
            scores = np.array([self.metric(self.y_, y_pred) for y_pred in y_preds])

            # Calculate regularizations and quality
            a = self.reg_a()
            b = self.reg_b()
            q = scores - (self.alpha * a) + (self.beta * b)

            # Update history
            self.all_scores.append(scores)
            self.all_qualities.append(q)

            # Sort model by obtained measure
            sorter = (-q).argsort()
            self.model_ = self.model_[sorter]
            scores = scores[sorter]
            q = q[sorter]

            # Selection
            if np.min(q) < 0:  # Deal with negative qualities
                p = q - np.min(q)
            else:
                p = q
            p[p == 0] = 0.0000001
            p = p[self.elite_limit :] / np.sum(p[self.elite_limit :])
            c = np.random.choice(
                range(self.elite_limit, self.pool_size),
                size=self.pool_size - self.elite_limit,
                p=p,
            )
            self.model_[self.elite_limit :] = self.model_[c]
            q[self.elite_limit :] = q[c]

        # Convert scores
        self.all_scores = np.array(self.all_scores)
        self.all_qualities = np.array(self.all_qualities)

    def reg_a(self):
        return np.sum(np.max(self.model_, axis=1), axis=1) / self.n_features

    def reg_b(self):
        d = np.zeros((self.pool_size)).astype(int)
        for e, ensemble in enumerate(self.model_):
            for i in range(self.ensemble_size):
                for j in range(i + 1, self.ensemble_size):
                    d[e] += len(np.bitwise_xor(ensemble[i], ensemble[j]).nonzero()[0])
        return d / binom(self.ensemble_size, 2) / self.n_features

    def get_model_probas(self):
        # Prepare structure
        probas = np.zeros(
            (self.pool_size, self.ensemble_size, self.n_objects, self.n_classes)
        )

        # Iterate ensembles
        for e, ensemble in enumerate(self.model_):
            for s, subspace in enumerate(ensemble):
                # Check if probing already done for subspace
                if tuple(subspace) not in self.probas_storage:
                    if hasattr(self, "coef_"):
                        self.probas_storage.update(
                            {
                                tuple(subspace): self._predict_proba_with_coef(
                                    self.X_, self.coef_ * subspace
                                )
                            }
                        )
                    if hasattr(self, "sigma_"):
                        self.probas_storage.update(
                            {
                                tuple(subspace): self._predict_proba_with_sigma(
                                    self.X_, self.sigma_ * subspace
                                )
                            }
                        )
                # Store
                probas[e, s] = self.probas_storage[tuple(subspace)]
        return probas

    def predict(self, X):
        if hasattr(self, "coef_"):
            predict_probas = np.array(
                [
                    self._predict_proba_with_coef(X, self.coef_ * subspace)
                    for subspace in self.model_[0]
                ]
            )
        if hasattr(self, "sigma_"):
            predict_probas = np.array(
                [
                    self._predict_proba_with_sigma(X, self.sigma_ * subspace)
                    for subspace in self.model_[0]
                ]
            )
        predict_probas = np.mean(predict_probas, axis=0)
        prediction = np.argmax(predict_probas, axis=1)
        return prediction

    def _predict_proba_with_coef(self, X, coef_):
        self.base_clf.coef_[0] = coef_
        predict_proba = self.base_clf.predict_proba(X)
        return predict_proba

    def _predict_proba_with_sigma(self, X, sigma_):
        sigma_[sigma_ == 0] = VAR
        self.base_clf.sigma_ = sigma_
        predict_proba = self.base_clf.predict_proba(X)
        return predict_proba

    def mutate(self):
        # Generate random mutation mask
        mask = (
            np.random.rand(self.pool_size, self.ensemble_size, self.n_features)
            < self.p_mutation
        )
        # Conserve elite
        mask[: self.elite_limit] = False

        # Mutate
        self.model_[mask] = np.invert(self.model_[mask])

    def cross(self):
        # Copy elite
        _elite = np.copy(self.model_[: self.elite_limit])

        # Match partners with given probability
        to_replace = np.where(
            np.random.rand(self.pool_size - self.elite_limit) < self.p_crossing
        )[0]
        partners = np.random.randint(self.pool_size, size=to_replace.size)

        # Generate crossing pattern
        pattern = (
            np.random.randint(
                2, size=(to_replace.size, self.ensemble_size, self.n_features)
            )
            == 1
        )

        # Perform all crossings
        for m, (i, j) in enumerate(zip(to_replace, partners)):
            self.model_[i, pattern[m]] = self.model_[j, pattern[m]]

        # Restore elite
        self.model_[: self.elite_limit] = _elite



from imblearn import under_sampling, over_sampling
from sklearn import base, model_selection, metrics, preprocessing
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats


class UMCE:
    """
    Shuffled K-fold undersampled ensemble.
    """

    def __init__(self, base_clf, dbname):
        self.base_clf = base_clf
        self.dbname = dbname
        self.clfs = []

    def fit(self, X_train, y_train):
        # Save X and y
        self.X_train = X_train
        self.y_train = y_train

        # Firstly we analyze the training set to find majority class and to
        # establish the imbalance ratio
        self.classes, c_counts = np.unique(y_train, return_counts=True)
        majority_c = 0 if c_counts[0] > c_counts[1] else 1
        minority_c = 1 - majority_c

        min_idx = np.where(y_train == minority_c)[0]
        maj_idx = np.where(y_train == majority_c)[0]

        # K is the imbalanced ratio round to int, being also a number of
        # ensemble members.
        imbalance_ratio = c_counts[majority_c] / c_counts[minority_c]
        self.k = int(np.around(imbalance_ratio))
        self.k = self.k if self.k > 2 else 2

        # We use k to KFold division of majority class
        self.clfs = []
        kf = model_selection.KFold(n_splits=self.k, shuffle=True)
        for _, index in kf.split(maj_idx):
            fold_idx = np.concatenate([min_idx, maj_idx[index]])
            X_train_f, y_train_f = X_train[fold_idx], y_train[fold_idx]

            clf = base.clone(self.base_clf)
            clf.fit(X_train_f, y_train_f)
            self.clfs.append(clf)

        # Add OS
        clf = base.clone(self.base_clf)
        os = over_sampling.RandomOverSampler()
        X_os, y_os = os.fit_sample(self.X_train, self.y_train)
        clf.fit(X_os, y_os)
        self.clfs.append(clf)

        # Calculate weights as balanced accuracy on whole set
        self.weights = np.array(
            [
                metrics.balanced_accuracy_score(self.y_train, clf.predict(self.X_train))
                for clf in self.clfs
            ]
        )

        scaler = preprocessing.MinMaxScaler()
        self.nweights = scaler.fit_transform(self.weights.reshape(-1, 1)).T[0]
        self.nweights += 0.01

    def test(self, X_test):
        self.X_test = X_test
        self.esc = ensemble_support_cube(self.clfs, X_test)

        # Analyze dependencies
        segments = list(range(len(self.clfs)))
        p_treshold = 0.01
        for i in range(self.k):
            a = self.esc[i, :, 0]
            for j in range(i + 1, self.k):
                b = self.esc[j, :, 0]
                p = stats.wilcoxon(a, b).pvalue
                c = (i, j)
                if p < p_treshold:
                    segments[i] = j

        # Reduced ESC and weights
        self.resc = np.array(
            [
                np.mean(
                    self.esc[segments == i, :, :]
                    * self.weights[segments == i, np.newaxis, np.newaxis],
                    axis=0,
                )
                for i in np.unique(segments)
            ]
        )
        self.rweights = np.array(
            [np.mean(self.weights[segments == i]) for i in np.unique(segments)]
        )

        scaler = preprocessing.MinMaxScaler()
        self.nrweights = scaler.fit_transform(self.rweights.reshape(-1, 1)).T[0]
        self.nrweights += 0.01

        # Contrasts
        self.contrast = np.abs(self.esc[:, :, 0] - self.esc[:, :, 1])
        self.rcontrast = np.abs(self.resc[:, :, 0] - self.resc[:, :, 1])

        # Calculate all the measures
        self.wesc = self.esc * self.weights[:, np.newaxis, np.newaxis]
        self.nwesc = self.esc * self.nweights[:, np.newaxis, np.newaxis]
        self.cwesc = self.wesc * self.contrast[:, :, np.newaxis]
        self.ncwesc = self.nwesc * self.contrast[:, :, np.newaxis]

        self.rwesc = self.resc * self.rweights[:, np.newaxis, np.newaxis]
        self.nrwesc = self.resc * self.nrweights[:, np.newaxis, np.newaxis]
        self.rcwesc = self.rwesc * self.rcontrast[:, :, np.newaxis]
        self.nrcwesc = self.nrwesc * self.rcontrast[:, :, np.newaxis]

    def decision_cube(self, reduced=False, mode="regular", os=False):
        esc_ = None
        if reduced:
            if mode == "regular":
                esc_ = self.esc
            elif mode == "weighted":
                esc_ = self.wesc
            elif mode == "cweighted":
                esc_ = self.cwesc
            elif mode == "nweighted":
                esc_ = self.nwesc
            elif mode == "ncweighted":
                esc_ = self.ncwesc
        else:
            if mode == "regular":
                esc_ = self.resc
            elif mode == "weighted":
                esc_ = self.rwesc
            elif mode == "cweighted":
                esc_ = self.rcwesc
            elif mode == "nweighted":
                esc_ = self.nrwesc
            elif mode == "ncweighted":
                esc_ = self.nrcwesc

        if os:
            return esc_
        else:
            return esc_[:-1, :, :]

    def predict(self, esc):
        # Basic soft voting
        return np.argmax(np.sum(esc, axis=0), axis=1)


def ensemble_support_cube(clfs, X):
    return np.array([clf.predict_proba(X) for clf in clfs])


def regular_bac(base_clf, X_train, y_train, X_test, y_test):
    regular_clf = base.clone(base_clf)
    regular_clf.fit(X_train, y_train)
    regular_pred = regular_clf.predict(X_test)
    return metrics.balanced_accuracy_score(y_test, regular_pred)


def us_os_bac(base_clf, X_train, y_train, X_test, y_test):
    us = under_sampling.RandomUnderSampler()
    os = over_sampling.RandomOverSampler()
    X_us, y_us = us.fit_sample(X_train, y_train)
    X_os, y_os = os.fit_sample(X_train, y_train)
    us_clf = base.clone(base_clf)
    os_clf = base.clone(base_clf)
    us_clf.fit(X_us, y_us)
    os_clf.fit(X_os, y_os)
    us_pred = us_clf.predict(X_test)
    os_pred = os_clf.predict(X_test)
    return (
        metrics.balanced_accuracy_score(y_test, us_pred),
        metrics.balanced_accuracy_score(y_test, os_pred),
    )



"""
Class for preparing datasets for the evaluator.
"""

# imports
import numpy as np
import os
import re


class Data():
    def __init__(self, selection, path="../../datasets/"):
        self.selection = selection
        self.path = path

    def load(self):
        """
        description

        Returns:
        datasets: dictionary containing dataset name and and a numpy array with
                  features and labels
                  "name" : numpy array (n_samples, n_features+1)
        """
        # load files
        files = self.__dir2files(path=self.path)
        datasets = {}

        if type(self.selection) == tuple:
            tag_filter = self.selection[1]
            for file in files:
                X, y, dbname, tags = self.__csv2Xy(file)
                if self.selection[0] == 'all':
                    is_good = all(elem in tags for elem in tag_filter)
                if self.selection[0] == 'any':
                    is_good = any(elem in tags for elem in tag_filter)
                if is_good:
                    datasets[dbname] = (X, y)

        # Load datasets by name
        elif type(self.selection) == list:
            for file in files:
                X, y, dbname, _ = self.__csv2Xy(file)
                is_good = dbname in self.selection
                if is_good:
                    datasets[dbname] = (X, y)
        else:
            print("Provide a list or a tuple")
        return datasets

    def multiply_dim(self, multiply=2):
        # load files
        files = self.__dir2files(path=self.path)
        datasets = {}

        # Load datasets by name
        if type(self.selection) == list:
            for file in files:
                X, y, dbname, _ = self.__csv2Xy(file)
                is_good = dbname in self.selection
                if is_good:
                    datasets[dbname + "_1"] = (X, y)
                    Xm = X
                    for i in range(multiply):
                        Xm = np.concatenate((Xm, X), axis=1)
                        dbname_m = dbname + "_" + str(i+2)
                        datasets[dbname_m] = (Xm, y)
        else:
            print("Provide a list of dataset names")
        return datasets

    def __tags4Xy(self, X, y):
        tags = []
        numberOfFeatures = X.shape[1]
        # numberOfSamples = len(y)
        numberOfClasses = len(np.unique(y))
        if numberOfClasses == 2:
            tags.append("binary")
        else:
            tags.append("multi-class")
        if numberOfFeatures >= 8:
            tags.append("multi-feature")

        # Calculate ratio
        ratio = [0.0] * numberOfClasses
        for y_ in y:
            ratio[y_] += 1
        ratio = [int(round(i / min(ratio))) for i in ratio]
        if max(ratio) > 3:
            tags.append("imbalanced")
        else:
            tags.append("balanced")
        return tags

    def __csv2Xy(self, path):
        ds = np.genfromtxt(path, delimiter=",")
        X = ds[:, :-1]
        y = ds[:, -1].astype(int)
        dbname = path.split("/")[-1].split(".")[0]
        tags = self.__tags4Xy(X, y)
        return X, y, dbname, tags
        # return datasets

    def __dir2files(self, path, extension="csv"):
        return [
            path + x
            for x in os.listdir(path)
            if re.match("^([a-zA-Z0-9-_])+\.%s$" % extension, x)
        ]



import numpy as np
from ..statistics import statistics
import warnings
from tabulate import tabulate


class PairedTests():
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def process(self, test_name, alpha=.05, m_fmt="%.3f", std_fmt=None,
                nc="---", db_fmt="%s", tablefmt="plain", **kwargs):
        try:
            test, tkwargs = statistics.IMPLEMENTED_TESTS[test_name]
            # Remove kwarg missmatches and update tkwargs
            missmatches = list(set(kwargs.keys()).difference(tkwargs.keys()))
            if len(missmatches) > 0:
                WARN_MESSAGE = '%s parameter(s) of %s are not supported and will be ignored' % (
                    ", ".join(missmatches), test_name)
                warnings.warn(WARN_MESSAGE)
                for key in missmatches:
                    del kwargs[key]
            tkwargs.update(kwargs)

            try:
                # Gather data
                scores = self.evaluator.scores
                mean_scores = self.evaluator.mean_scores
                stds = self.evaluator.stds
                metrics = list(self.evaluator.metrics.keys())
                clfs = list(self.evaluator.clfs.keys())
                datasets = list(self.evaluator.datasets.keys())
                n_clfs = len(clfs)

                # Perform tests
                tables = {}
                for m_idx, m_name in enumerate(metrics):
                    # Prepare storage for table
                    t = []

                    for db_idx, db_name in enumerate(datasets):
                        # Row with mean scores
                        t.append([db_fmt % db_name] + [m_fmt %
                                                       v for v in
                                                       mean_scores[db_idx, :,
                                                                   m_idx]])

                        # Row with std
                        if std_fmt:
                            t.append([''] + [std_fmt %
                                             v for v in
                                             stds[db_idx, :, m_idx]])
                        # Calculate T and p
                        T, p = np.array(
                            [[test(scores[db_idx, i, :, m_idx],
                                   scores[db_idx, j, :, m_idx],
                                   **tkwargs)
                              for i in range(n_clfs)]
                             for j in range(n_clfs)]
                        ).swapaxes(0, 2)
                        _ = np.where((p < alpha) * (T > 0))
                        conclusions = [list(1 + _[1][_[0] == i])
                                       for i in range(n_clfs)]

                        # Row with conclusions
                        t.append([''] + [", ".join(["%i" % i for i in c])
                                         if len(c) > 0 else nc
                                         for c in conclusions])

                    # Store formatted table
                    tables.update({m_name: tabulate(
                        t, headers=['DATASET'] + clfs, tablefmt=tablefmt)})

                return tables
            except AttributeError:
                WARN_MESSAGE = "Call evaluator's score method first."
                raise Exception(WARN_MESSAGE)
        except KeyError:
            WARN_MESSAGE = '%s is not a supported test. Use one of: %s.' % (
                test_name, ", ".join(list(statistics.IMPLEMENTED_TESTS.keys()))
            )
            warnings.warn(WARN_MESSAGE)



"""
description
"""

# imports
from sklearn.datasets import make_classification
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import accuracy_score
import weles as ws


def dataset():
    return make_classification(random_state=1410)


def test_evaluator():
    metrics = {"accuracy": accuracy_score}
    clfs = {"GNB": GaussianNB(),
            "KNC": KNeighborsClassifier(),
            "DTC": DecisionTreeClassifier()}

    ev = ws.evaluation.Evaluator(datasets={"dup": dataset()})
    ev.process(clfs=clfs, verbose=True)
    ev.score(metrics=metrics, verbose=True)



from sklearn.base import ClassifierMixin, BaseEstimator, clone
from sklearn.cluster import KMeans
import numpy as np


class Cluens(ClassifierMixin, BaseEstimator):
    def __init__(self, base_estimator, n_estimators=3):
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators

    def fit(self, X, y):
        self.classes = np.unique(y)
        self.clstrs = [
            KMeans(n_clusters=self.n_estimators).fit(X[y == c]) for c in self.classes
        ]

        self.estimators_ = []
        for clstr in self.clstrs:
            y_pred = clstr.predict(X)
            for cls in range(self.n_estimators):
                mask = y_pred == cls
                if len(np.unique(y[mask])) == 2:
                    clf = clone(self.base_estimator).fit(X[mask], y[mask])
                    self.estimators_.append(clf)

        return self



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin


class LinearClassifier(BaseEstimator, ClassifierMixin):
    def fit(self, X, y):
        p = np.array([np.mean(X[y == c], axis=0) for c in np.unique(y)])
        self.b = np.mean(p, axis=0)
        self.w = p[1] - self.b
        return self

    def decision_function(self, X):
        return (X - self.b).dot(self.w)

    def predict(self, X):
        return self.decision_function(X) > 0



# Empty
from ._version import __version__

from . import classifiers
from . import ensembles
from . import meta
from . import optimizers
from . import statistics
from . import evaluation
from . import utils
from . import metrics

__all__ = [
    "classifiers",
    "ensembles",
    "meta",
    "optimizers",
    "statistics",
    "evaluation",
    "utils",
    "metrics",
    "__version__",
]



#! /usr/bin/env python
from __future__ import absolute_import

import codecs
import os

from setuptools import find_packages, setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# get __version__ from _version.py
ver_file = os.path.join("weles", "_version.py")
with open(ver_file) as f:
    exec(f.read())

DISTNAME = "weles"
DESCRIPTION = "Supplementary scikit-learn package with methods developed in W4K2."
MAINTAINER = "P. Ksieniewicz"
MAINTAINER_EMAIL = "pawel.ksieniewicz@pwr.edu.pl"
URL = "https://github.com/w4k2/weles"
LICENSE = "GPL-3.0"
DOWNLOAD_URL = "https://github.com/w4k2/weles"
VERSION = __version__
INSTALL_REQUIRES = ["numpy", "scipy", "scikit-learn", "imblearn", "tabulate", "tqdm"]
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    packages=find_packages(),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=CLASSIFIERS,
)



"""
Absolutnie bez sensu. Powiela verbose z Evaluator.score().
"""

import numpy as np
from scipy.stats import rankdata
from tabulate import tabulate
from scipy.stats import wilcoxon


class Analyzer():
    def __init__(self, scores, clfs, metrics, datasets):
        self.scores = scores
        self.clfs = clfs
        self.metrics = metrics
        self.datasets = datasets

    def analyze(self, alpha=.05, stat_rank=wilcoxon):
        for m, metric in enumerate(self.metrics):
            print("################ ", metric, " ################")
            scores = self.scores[:,:,m]

            names_column = np.array(list(self.datasets.keys())).reshape(len(self.datasets), -1)
            scores_table = np.concatenate((names_column, scores), axis=1)
            print(tabulate(scores_table, headers=self.clfs.keys(), floatfmt=".3f"), scores_table.shape)

            # ranks
            ranks = []
            for row in scores:
                ranks.append(rankdata(row).tolist())
            ranks = np.array(ranks)
            mean_ranks = np.mean(ranks, axis=0)

            # strasznie głupi test globalny do texa
            # p_value = np.zeros((len(self.clfs), len(self.clfs)))
            #
            # for i in range(len(self.clfs)):
            #     for j in range(len(self.clfs)):
            #         _, p_value[i, j] = stat_rank(ranks.T[i], ranks.T[j], zero_method="zsplit")
            # dependency = p_value>alpha
            #
            # text = "\\ "
            # for i in range(mean_ranks.shape[0]):
            #     text += "& "
            #     a = np.where(dependency[i] == 0)[0]
            #
            #     for value in a:
            #         if mean_ranks[i] < mean_ranks[value]:
            #             a = a[a != value]
            #
            #     if a.size == mean_ranks.shape[0]-1:
            #         text += "$_{all}$"
            #     elif a.size == 0:
            #         text += "$_{-}$"
            #     else:
            #         a += 1
            #         text +=  "$_{" + ", ".join(["%i" % i for i in a]) + "}$"
            #
            # text += "\\\\\n"

            print("################ Mean ranks ################")
            print(tabulate(mean_ranks[np.newaxis,:], headers=self.clfs.keys(), floatfmt=".3f" ))



from sklearn.datasets import make_classification
import weles as ws


def dataset():
    return make_classification(random_state=1410)


def test_linear_classifier():
    clf = ws.classifiers.LinearClassifier()
    X, y = dataset()
    clf.fit(X, y)
    y_pred = clf.predict(X)


def test_exposer_classifier():
    clf = ws.classifiers.ExposerClassifier()
    X, y = dataset()
    clf.fit(X, y)
    y_pred = clf.predict(X)

    for i in range(5):
        clf.partial_fit(X, y)


def test_ssgnb():
    clf = ws.classifiers.SSGNB()
    X, y = dataset()
    clf.fit(X, y)
    y_pred = clf.predict(X)
    subspaces = [[0, 1, 2], [3], [7, 9]]
    y_preds = clf.predict(X, subspaces)
